/**
 * Copyright (c) 2004 Brian A. Weaver <weave@spellweaver.org>
 * All Rights Reserverd
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 */
#include <sys/types.h>
#include <string.h>
#include <pwd.h>
#include <grp.h>
#include <syslog.h>

#define PAM_SM_AUTH
#include <security/pam_modules.h>

#define THIS_MODULE_NAME "pam(lockout)"

/**
 * Define the beginning of our options keys. This is the users
 * and groups that we wish to lockout of the system
 */
#define USER_ARGKEY "user="
#define USER_ARGLEN 5
#define GROUP_ARGKEY "group="
#define GROUP_ARGLEN 6

/**
 * This is the authentication portion of the lockout module. It will
 * determine if the user attempting to access the system should be
 * locked out by it's primary user or group.
 *
 */
PAM_EXTERN
int pam_sm_authenticate(pam_handle_t *pamh,
						int flags,
						int argc,
						const char **argv)
{
	int 			iRC;		/* return code */
	int 			result;		/* authentication result */
	const char *	user;		/* user to validate */
	struct passwd * pwent;		/* user's password entry */
	struct group *  grent;		/* group entry */
	uid_t			thisuser;	/* the id of the autentication user */
	gid_t			thisgroup;	/* the group of the user */
	int 			ndx;		/* a counter index */

	/*
	 * get the user name
	 */
	result = PAM_AUTH_ERR;
	user   = NULL;
	iRC    = pam_get_user(pamh, &user, NULL);
	if(iRC == PAM_SUCCESS && user != NULL)
	{
		/*
		 * load up the password entry for this user.
		 */
		pwent = getpwnam(user);
		if(pwent == NULL)
		{
			result = PAM_USER_UNKNOWN;
			goto onerr;
		}
		thisuser  = pwent->pw_uid;	/* save the user */
		thisgroup = pwent->pw_gid;	/* save the group */

		/*
		 * set the result to OK and then see if we find a reason to 
		 * fail this user
		 */
		result = PAM_SUCCESS;

		/*
		 * Check the user against locked out users and groups
		 */
		for(ndx = 0; ndx < argc; ndx++)
		{

			if(strncmp(USER_ARGKEY, argv[ndx], USER_ARGLEN) == 0 && *(argv[ndx] + USER_ARGLEN) != '\0')
			{
				/**
				 * this is the simple check. Pull the entry for the command line
				 * user and then compare the user ids
				 */
				pwent = getpwnam(argv[ndx] + USER_ARGLEN);
				if(pwent != NULL && pwent->pw_uid == thisuser)
				{
					result = PAM_AUTH_ERR;
					goto onerr;
				}
			}
			else if(strncmp(GROUP_ARGKEY, argv[ndx], GROUP_ARGLEN) == 0 && *(argv[ndx] + GROUP_ARGLEN) != '\0')
			{
				/**
				 * preform the simple check first, which is the group of the
				 * authenticating user verses the group on the cmd line.
				 */
				grent = getgrnam(argv[ndx] + GROUP_ARGLEN);
				if(grent != NULL && grent->gr_gid == thisgroup)
				{
					result = PAM_AUTH_ERR;
					goto onerr;
				}
				else if(grent != NULL && grent->gr_mem != NULL) /* enumerate the members */
				{
					char ** member = grent->gr_mem;
					while(member != NULL && *member != NULL)
					{
						if(strcmp(user, *member) == 0)
						{
							result = PAM_AUTH_ERR;
							goto onerr;
						}
						++member;
					}
				}
			}
			else /* process unknown options */
			{
				openlog(THIS_MODULE_NAME, LOG_NDELAY, LOG_AUTHPRIV);
				syslog(LOG_ERR, "unrecognized module option %s", argv[ndx]);
				closelog();
			}

		} /* end looping over command line options */

	} 
	else /* could not recover name, log the error */
	{
		openlog(THIS_MODULE_NAME, LOG_NDELAY, LOG_AUTHPRIV);
		syslog(LOG_ERR, "failed to get user name from pam libraries");
		closelog();
	}

onerr:
	/**
	 * clean up if necessary. Currently there is nothing to close or cleanup
	 * Also, log if there is a failure.
	 */
	openlog(THIS_MODULE_NAME, LOG_NDELAY, LOG_AUTHPRIV);
	if(result == PAM_AUTH_ERR && user != NULL)
	{
		syslog(LOG_NOTICE, "user %s denied access due to lockout", user);
	}
	else if(result == PAM_USER_UNKNOWN && user != NULL)
	{
		syslog(LOG_NOTICE, "user %s is not a valid user", user);
	}

	return result;
}

/**
 * We don't do credential management, so this always succeeds.
 */
PAM_EXTERN
int pam_sm_setcred(pam_handle_t *pamh, int flags, int argc, const char **argv)
{
	return PAM_SUCCESS;
}

/*
 * static module data
 */
#ifdef PAM_STATIC
struct pam_module _pam_deny_modstruct = {
	"pam_lockout",
	pam_sm_authenticate,
	pam_sm_setcred,
	NULL,
	NULL,
	NULL,
	NULL
};
#endif /* PAM_STATIC */

/*
 * vim:formatoptions-=r:shiftwidth=4:tabstop=4
 */
